@@ -12,6 +12,51 @@ import sys
 import traceback
 import requests
 import re
+# from .HTTPProxyDigestAuth import HTTPProxyDigestAuth
+
+logging.basicConfig()
+CLIENT_APPLICATION = 'qiskit-api-py'
+
+
+def get_job_url(config, hub, group, project):
+    """
+    Util method to get job url
+    """
+    if ((config is not None) and ('hub' in config) and (hub is None)):
+        hub = config["hub"]
+    if ((config is not None) and ('group' in config) and (group is None)):
+        group = config["group"]
+    if ((config is not None) and ('project' in config) and (project is None)):
+        project = config["project"]
+    if ((hub is not None) and (group is not None) and (project is not None)):
+        return '/Network/{}/Groups/{}/Projects/{}/jobs'.format(hub, group, project)
+    return '/Jobs'
+
+
+def get_backend_stats_url(config, hub, backend_type):
+    """
+    Util method to get backend stats url
+    """
+    if ((config is not None) and ('hub' in config) and (hub is None)):
+        hub = config["hub"]
+    if (hub is not None):
+        return '/Network/{}/devices/{}'.format(hub, backend_type)
+    return '/Backends/{}'.format(backend_type)
+
+
+def get_backend_url(config, hub, group, project):
+    """
+    Util method to get backend url
+    """
+    if ((config is not None) and ('hub' in config) and (hub is None)):
+        hub = config["hub"]
+    if ((config is not None) and ('group' in config) and (group is None)):
+        group = config["group"]
+    if ((config is not None) and ('project' in config) and (project is None)):
+        project = config["project"]
+    if ((hub is not None) and (group is not None) and (project is not None)):
+        return '/Network/{}/Groups/{}/Projects/{}/devices'.format(hub, group, project)
+    return '/Backends'
 
 
 class _Credentials(object):
@@ -20,33 +65,93 @@ class _Credentials(object):
     """
     config_base = {'url': 'https://quantumexperience.ng.bluemix.net/api'}
 
-    def __init__(self, token, config=None, verify=True):
+    def __init__(self, token, config=None, verify=True, proxies=None):
         self.token_unique = token
         self.verify = verify
+        self.config = config
+        self.proxies = proxies
         if not verify:
             import requests.packages.urllib3 as urllib3
             urllib3.disable_warnings()
             print('-- Ignoring SSL errors.  This is not recommended --')
-        if config and config.get('url', None):
-            self.config = config
-        else:
+        if self.config and ("url" not in self.config):
+            self.config["url"] = self.config_base["url"]
+        elif not self.config:
             self.config = self.config_base
 
         self.data_credentials = {}
         if token:
-            self.obtain_token()
+            self.obtain_token(config=self.config)
+        else:
+            access_token = self.config.get('access_token', None)
+            if access_token:
+                user_id = self.config.get('user_id', None)
+                if access_token:
+                    self.set_token(access_token)
+                if user_id:
+                    self.set_user_id(user_id)
+            else:
+                self.obtain_token(config=self.config)
 
-    def obtain_token(self):
+    def obtain_token(self, config=None):
         """Obtain the token to access to QX Platform.
 
         Raises:
-            CredentialsError: when token is invalid.
+            CredentialsError: when token is invalid or the user has not
+                accepted the license.
+            ApiError: when the response from the server couldn't be parsed.
         """
-        self.data_credentials = requests.post(str(self.config.get('url') +
-                                                  "/users/loginWithToken"),
-                                              data={'apiToken':
-                                                    self.token_unique},
-                                              verify=self.verify).json()
+        client_application = CLIENT_APPLICATION
+        if self.config and ("client_application" in self.config):
+            client_application += ':' + self.config["client_application"]
+        headers = {'x-qx-client-application': client_application}
+        if self.token_unique:
+            try:
+                response = requests.post(str(self.config.get('url') +
+                                             "/users/loginWithToken"),
+                                         data={'apiToken': self.token_unique},
+                                         verify=self.verify,
+                                         headers=headers,
+                                         proxies=self.proxies)
+            except requests.RequestException as e:
+                raise ApiError('error during login: %s' % str(e))
+        elif config and ("email" in config) and ("password" in config):
+            email = config.get('email', None)
+            password = config.get('password', None)
+            credentials = {
+                'email': email,
+                'password': password
+            }
+            try:
+                response = requests.post(str(self.config.get('url') +
+                                             "/users/login"),
+                                         data=credentials,
+                                         verify=self.verify,
+                                         headers=headers,
+                                         proxies=self.proxies)
+            except requests.RequestException as e:
+                raise ApiError('error during login: %s' % str(e))
+        else:
+            raise CredentialsError('invalid token')
+
+        if response.status_code == 401:
+            error_message = None
+            try:
+                # For 401: ACCEPT_LICENSE_REQUIRED, a detailed message is
+                # present in the response and passed to the exception.
+                error_message = response.json()['error']['message']
+            except:
+                pass
+
+            if error_message:
+                raise CredentialsError('error during login: %s' % error_message)
+            else:
+                raise CredentialsError('invalid token')
+        try:
+            response.raise_for_status()
+            self.data_credentials = response.json()
+        except (requests.HTTPError, json.decoder.JSONDecodeError) as e:
+            raise ApiError('error during login: %s' % str(e))
 
         if self.get_token() is None:
             raise CredentialsError('invalid token')
@@ -89,7 +194,16 @@ class _Request(object):
     def __init__(self, token, config=None, verify=True, retries=5,
                  timeout_interval=1.0):
         self.verify = verify
-        self.credential = _Credentials(token, config, verify)
+        self.client_application = CLIENT_APPLICATION
+        self.config = config
+        self.proxies = None
+        if ((config is not None) and ('proxies' in config) and
+           ('urls' in config['proxies'])):
+            self.proxies = self.config['proxies']['urls']
+        if self.config and ("client_application" in self.config):
+            self.client_application += ':' + self.config["client_application"]
+        self.credential = _Credentials(token, self.config, verify,
+                                       proxies=self.proxies)
         self.log = logging.getLogger(__name__)
         if not isinstance(retries, int):
             raise TypeError('post retries must be positive integer')
@@ -105,7 +219,7 @@ class _Request(object):
         Check is the user's token is valid
         """
         if respond.status_code == 401:
-            self.credential.obtain_token()
+            self.credential.obtain_token(config=self.config)
             return False
         return True
 
@@ -113,19 +227,61 @@ class _Request(object):
         """
         POST Method Wrapper of the REST API
         """
+        self.result = None
         data = data or {}
-        headers = {'Content-Type': 'application/json'}
+        headers = {'Content-Type': 'application/json',
+                   'x-qx-client-application': self.client_application}
         url = str(self.credential.config['url'] + path + '?access_token=' +
                   self.credential.get_token() + params)
         retries = self.retries
         while retries > 0:
             respond = requests.post(url, data=data, headers=headers,
-                                    verify=self.verify)
+                                    verify=self.verify, proxies=self.proxies)
             if not self.check_token(respond):
                 respond = requests.post(url, data=data, headers=headers,
-                                        verify=self.verify)
+                                        verify=self.verify,
+                                        proxies=self.proxies)
+
             if self._response_good(respond):
-                return self.result
+                if self.result:
+                    return self.result
+                elif retries < 2:
+                    return respond.json()
+                else:
+                    retries -= 1
+            else:
+                retries -= 1
+                time.sleep(self.timeout_interval)
+
+        # timed out
+        raise ApiError(usr_msg='Failed to get proper ' +
+                       'response from backend.')
+
+    def put(self, path, params='', data=None):
+        """
+        PUT Method Wrapper of the REST API
+        """
+        self.result = None
+        data = data or {}
+        headers = {'Content-Type': 'application/json',
+                   'x-qx-client-application': self.client_application}
+        url = str(self.credential.config['url'] + path + '?access_token=' +
+                  self.credential.get_token() + params)
+        retries = self.retries
+        while retries > 0:
+            respond = requests.put(url, data=data, headers=headers,
+                                   verify=self.verify, proxies=self.proxies)
+            if not self.check_token(respond):
+                respond = requests.put(url, data=data, headers=headers,
+                                       verify=self.verify,
+                                       proxies=self.proxies)
+            if self._response_good(respond):
+                if self.result:
+                    return self.result
+                elif retries < 2:
+                    return respond.json()
+                else:
+                    retries -= 1
             else:
                 retries -= 1
                 time.sleep(self.timeout_interval)
@@ -137,6 +293,7 @@ class _Request(object):
         """
         GET Method Wrapper of the REST API
         """
+        self.result = None
         access_token = ''
         if with_token:
             access_token = self.credential.get_token() or ''
@@ -144,12 +301,20 @@ class _Request(object):
                 access_token = '?access_token=' + str(access_token)
         url = self.credential.config['url'] + path + access_token + params
         retries = self.retries
-        while retries > 0: # Repeat until no error
-            respond = requests.get(url, verify=self.verify)
+        headers = {'x-qx-client-application': self.client_application}
+        while retries > 0:  # Repeat until no error
+            respond = requests.get(url, verify=self.verify, headers=headers,
+                                   proxies=self.proxies)
             if not self.check_token(respond):
-                respond = requests.get(url, verify=self.verify)
+                respond = requests.get(url, verify=self.verify,
+                                       headers=headers, proxies=self.proxies)
             if self._response_good(respond):
-                return self.result
+                if self.result:
+                    return self.result
+                elif retries < 2:
+                    return respond.json()
+                else:
+                    retries -= 1
             else:
                 retries -= 1
                 time.sleep(self.timeout_interval)
@@ -176,7 +341,11 @@ class _Request(object):
                 respond.text))
             return self._parse_response(respond)
         try:
-            self.result = respond.json()
+            if (str(respond.headers['content-type']).startswith("text/html;")):
+                self.result = respond.text
+                return True
+            else:
+                self.result = respond.json()
         except (json.JSONDecodeError, ValueError):
             usr_msg = 'device server returned unexpected http response'
             dev_msg = usr_msg + ': ' + respond.text
@@ -232,6 +401,7 @@ class IBMQuantumExperience(object):
 
     def __init__(self, token=None, config=None, verify=True):
         """ If verify is set to false, ignore SSL certificate errors """
+        self.config = config
         self.req = _Request(token, config=config, verify=verify)
 
     def _check_backend(self, backend, endpoint):
@@ -304,7 +474,7 @@ class IBMQuantumExperience(object):
 
     def get_result_from_execution(self, id_execution, access_token=None, user_id=None):
         """
-        Get the result of a execution, byt the execution id
+        Get the result of a execution, by the execution id
         """
         if access_token:
             self.req.credential.set_token(access_token)
@@ -458,7 +628,8 @@ class IBMQuantumExperience(object):
             return respond
 
     def run_job(self, qasms, backend='simulator', shots=1,
-                max_credits=3, seed=None, access_token=None, user_id=None):
+                max_credits=None, seed=None, hub=None, group=None,
+                project=None, hpc=None, access_token=None, user_id=None):
         """
         Execute a job
         """
@@ -473,8 +644,12 @@ class IBMQuantumExperience(object):
             qasm['qasm'] = qasm['qasm'].replace('OPENQASM 2.0;', '')
         data = {'qasms': qasms,
                 'shots': shots,
-                'maxCredits': max_credits,
                 'backend': {}}
+        if max_credits:
+            data['maxCredits'] = max_credits
+
+        if hpc:
+            data['hpc'] = hpc
 
         backend_type = self._check_backend(backend, 'job')
 
@@ -487,10 +662,15 @@ class IBMQuantumExperience(object):
             return {"error": "Not seed allowed. Max 10 digits."}
 
         data['backend']['name'] = backend_type
-        job = self.req.post('/Jobs', data=json.dumps(data))
+
+        url = get_job_url(self.config, hub, group, project)
+
+        job = self.req.post(url, data=json.dumps(data))
+
         return job
 
-    def get_job(self, id_job, access_token=None, user_id=None):
+    def get_job(self, id_job, hub=None, group=None, project=None,
+                access_token=None, user_id=None):
         """
         Get the information about a job, by its id
         """
@@ -508,7 +688,12 @@ class IBMQuantumExperience(object):
             respond["status"] = 'Error'
             respond["error"] = "Job ID not specified"
             return respond
-        job = self.req.get('/Jobs/' + id_job)
+
+        url = get_job_url(self.config, hub, group, project)
+
+        url += '/' + id_job
+
+        job = self.req.get(url)
 
         # To remove result object and add the attributes to data object
         if 'qasms' in job:
@@ -535,7 +720,7 @@ class IBMQuantumExperience(object):
         jobs = self.req.get('/Jobs', '&filter={"limit":' + str(limit) + '}')
         return jobs
 
-    def backend_status(self, backend='ibmqx2', access_token=None, user_id=None):
+    def backend_status(self, backend='ibmqx4', access_token=None, user_id=None):
         """
         Get the status of a chip
         """
@@ -555,10 +740,14 @@ class IBMQuantumExperience(object):
             ret['available'] = bool(status['state'])
         if 'busy' in status:
             ret['busy'] = bool(status['busy'])
+        if 'lengthQueue' in status:
+            ret['pending_jobs'] = status['lengthQueue']
+        
+        ret['backend'] = backend_type
 
         return ret
 
-    def backend_calibration(self, backend='ibmqx2', access_token=None, user_id=None):
+    def backend_calibration(self, backend='ibmqx4', hub=None, access_token=None, user_id=None):
         """
         Get the calibration of a real chip
         """
@@ -580,11 +769,13 @@ class IBMQuantumExperience(object):
             ret["calibrations"] = None
             return ret
 
-        ret = self.req.get('/Backends/' + backend_type + '/calibration')
+        url = get_backend_stats_url(self.config, hub, backend_type)
+
+        ret = self.req.get(url + '/calibration')
         ret["backend"] = backend_type
         return ret
 
-    def backend_parameters(self, backend='ibmqx2', access_token=None, user_id=None):
+    def backend_parameters(self, backend='ibmqx4', hub=None, access_token=None, user_id=None):
         """
         Get the parameters of calibration of a real chip
         """
@@ -606,11 +797,13 @@ class IBMQuantumExperience(object):
             ret["parameters"] = None
             return ret
 
-        ret = self.req.get('/Backends/' + backend_type + '/parameters')
+        url = get_backend_stats_url(self.config, hub, backend_type)
+
+        ret = self.req.get(url + '/parameters')
         ret["backend"] = backend_type
         return ret
 
-    def available_backends(self, access_token=None, user_id=None):
+    def available_backends(self, hub=None, group=None, project=None, access_token=None, user_id=None):
         """
         Get the backends available to use in the QX Platform
         """
@@ -621,7 +814,13 @@ class IBMQuantumExperience(object):
         if not self.check_credentials():
             raise CredentialsError('credentials invalid')
         else:
-            return [backend for backend in self.req.get('/Backends')
+
+            url = get_backend_url(self.config, hub, group, project)
+
+            ret = self.req.get(url)
+            if (ret is not None) and (isinstance(ret, dict)):
+                return []
+            return [backend for backend in ret
                     if backend.get('status') == 'on']
 
     def available_backend_simulators(self, access_token=None, user_id=None):
@@ -635,13 +834,16 @@ class IBMQuantumExperience(object):
         if not self.check_credentials():
             raise CredentialsError('credentials invalid')
         else:
-            return [backend for backend in self.req.get('/Backends')
+            ret = self.req.get('/Backends')
+            if (ret is not None) and (isinstance(ret, dict)):
+                return []
+            return [backend for backend in ret
                     if backend.get('status') == 'on' and
                     backend.get('simulator') is True]
 
     def get_my_credits(self, access_token=None, user_id=None):
         """
-        Get the the credits by user to use in the QX Platform
+        Get the credits by user to use in the QX Platform
         """
         if access_token:
             self.req.credential.set_token(access_token)
@@ -660,6 +862,12 @@ class IBMQuantumExperience(object):
                 return user_data["credit"]
             return {}
 
+    def api_version(self):
+        """
+        Get the API Version of the QX Platform
+        """
+        return self.req.get('/version')
+
 
 class ApiError(Exception):
     """
